# Copyright MIT MRC 2013

#!/usr/bin/env python
""" daikon.py: Extract hachoir information in daikon classes """

__author__  = "Stelios Sidiroglou-Douskos <stelios@csail.mit.edu>"
__author__  = "Fan Long <fanl@csail.mit.edu>"
__author__  = "Neil Fitzgerald <neilfitz@mit.edu>"
__licence__ = "MIT"

# System
#import  pdb
import numpy as np
import ac2_logging
import logging
import random

import os
os.sys.path.append("/home/mrc/comfortfuzz")

from breaker import *
from breaker.arithmetic_utils import *
from breaker.dumb_fuzz import *
from constraint_breaker import breaker_solve

comparability_keys = {}

log = ac2_logging.logger("daikon-learning", loglevel=logging.DEBUG)

def is_number(s):
    try:
        float(s)
        return True
    except ValueError:
        return False

def get_random_key(maxint = 100000):
    while True:
        key = random.randrange(0,maxint)
        if key not in comparability_keys:
            comparability_keys[key] = 1
            return key

class DerivedVariable:
    def __init__(self, name):
        self.name           =   name
        self.iz3_dict       =   {}


class PPDecl:
    def __init__(self, name, random_comparability=True):
        self.name                 = name + ":::"
        self.variables            = {}
        self.files                = []
        self.random_comparability = random_comparability
        self.derived_variables    = {}

    def set_random_comparability(self, c):
        self.random_comparability = c

    def add_current_file(self, filename):
        """ Let PPDecl know about which files you have processed """
        self.files.append(filename)

    def ppt_decl(self):
        return "ppt %s" %(self.name)

    def type_decl(self):
        return "ppt-type object"

    def get_unique_variable_name(self, field, path):
        """
        DEPRECATED

        Create unique variable name based on path and descriptions when needed.

        When parsing JPG files, we got an assertion because we had more than one type
        per daikon variable. This was because of the entry fields in the EXIF format; they use
        description tags to differentiate themselves but otherwise look like arrays of entries.
        """
        display = field.display
        if not is_number(display):
            d = "".join([s for s in display.splitlines(True) if s.strip("\r\n")])
            d = "".join([s for s in d.splitlines(True) if s.strip("\r")])
            d = "".join([s for s in d.splitlines(True) if s.strip("\n")])
            d = d.replace("\\", "")
            #d = d.encode("utf-8")
            varname = path + "{" + d.replace(" ", "\_") + "}"
            print varname
        else:
            varname = path + "{blah}"
        return varname

    def add_variable(self, filename, field, path):
        varname = path
        if varname not in self.variables:
            var                         = VarDecl(varname, self.random_comparability)
            var.add_record(filename, field)
            self.variables[varname]     = var
        else:
            var = self.variables[varname]
            var.add_record(filename, field)

    def get_declarations(self):
        lines = [self.ppt_decl(), self.type_decl()]
        lines.extend(var.get_daikon_variable() for var in self.variables.values())
        return u"\n".join([line for line in lines if line != None])

    def get_trace(self):
        lines = []
        for filename in self.files:
            # Start by outputing declaration
            lines.extend("\n")
            lines.extend([self.name])

            for var in self.variables.values():
                if filename in var.records:
                    if len(var.records[filename]) == 1:
                        r   = var.records[filename][0]
                        if var.is_array:
                            val = "[%s]" %(str(r.value))
                        else:
                            val =  str(r.value)
                    else:
                        values  = []
                        for r in var.records[filename]:
                            values.append(str(r.value))
                        val     =  u"[%s]"%(" ".join(unicode(e) for e in values))
                    r_lines     = [var.get_sanitized_name(), val , "1"]
                    lines.extend(r_lines)
                else:
                    nan = "nonsensical"
                    #if var.is_array: nan = "[" + nan + "]"
                    r_lines = [var.get_sanitized_name(), nan, "2"]
                    lines.extend(r_lines)
            lines.extend("")
        return "\n".join([line for line in lines if line != None])

    def get_decl_version(self):
        return "decl-version 2.0\n"

    def write_to_file(self, filename):
        import codecs
        output_file = codecs.open(filename, "w", "utf-8")
        output_file.write(self.get_decl_version())
        output_file.write(self.get_declarations())
        output_file.write("\n")
        output_file.write(self.get_trace())
        output_file.close()

    def __str__(self):
        lines = [self.ppt_decl(), self.type_decl()]
        lines.extend(str(var) for var in self.variables.values())
        return "\n".join([line for line in lines if line != None])

    def __repr__(self):
        return self.__str__()

class VarDecl:
    def __init__(self, name, random_comparability=True):
        # Variable name
        self.name                                       = name
        # List of records (values seen for variable)
        self.records                                    = {}
        # Array flag
        self.is_array                                   = False
        # Flag for setting comparability fields
        self.random_comparability                       = random_comparability
        # Integer or string constraint class
        self.constraint_type                            = None

    def get_sanitized_name(self):
        return self.name.replace("\_", " ")

    def add_record(self, filename, field):
        rec = Record(field)

        if filename in self.records:
            record_list             = self.records[filename]
            record_list.append(rec)
        else:
            record_list             = []
            record_list.append(rec)
            self.records[filename]  = record_list

    def add_constraint_type(self):
        vtype = self.get_type_from_recs()
        if vtype in IntegerConstraint.type_map.keys():
            self.constraint_type = IntegerConstraint(vtype, self.get_values())
        elif vtype in StringConstraint.string_types:
            self.constraint_type = StringConstraint(vtype, self.get_values())

    def var_name(self):
        #return "variable %s" %self.name.encode("utf-8")
        return "variable %s" %self.name

    def var_kind(self):
        return "var-kind variable"

    def dec_type(self):
        #rec_type = [HachoirToInternalMap.type_map[y.original_type] for x in self.records.values() for y in x]
        #rec_set  = set(rec_type)
        #assert(len(rec_set) == 1), "%s" %rec_set
        #return "dec-type %s" %rec_set.pop()
        return "dec-type %s" %self.get_type_from_recs()

    def rep_type(self):
        return  "rep-type %s" %(self.get_daikon_type_from_recs())

    def enclosing_var(self):
        return "enclosing-var this"

    def comparability(self):
        """ Daikon comparability """
        if self.random_comparability:
            return "comparability %d" %get_random_key()
        else:
            return "comparability -1"

    def get_values(self):
        values = []
        for value in self.records.values():
            for record in value:
                values.append(record.value)
        return values

    def get_values_histogram(self):
        from collections import Counter
        return Counter(self.get_values())

    def get_original_type_from_recs(self):
        """ Get the variable type by sorting through the records """
        rec_type = []
        for rec_value in self.records.values():
            for rv in rec_value:
                rec_type.append(str(rv.original_type))
        rec_set = set(rec_type)
        return rec_set.pop()

    def get_type_from_recs(self):
        """ Get the variable type by sorting through the records """
        rec_type = []
        for rec_value in self.records.values():
            for rv in rec_value:
                rec_type.append(str(rv.type))
        rec_set = set(rec_type)
        return rec_set.pop()

    def get_daikon_type_from_recs(self):
        """ This is ugly """
        rec_type = []

        for r in self.records.values():
            if len(r) > 1: self.is_array = True
            for each in r:
                rec_type.append(each.get_daikon_type_name())

        rec_set = set(rec_type)
        if len(rec_set) > 1:
            print self.var_name()
            for r in self.records.values():
                for e in r:
                    print e.path, e.value, e.type
            print rec_set
            print self.var_name()
            assert False, "We are getting >1 type per record"

        type_name = rec_set.pop()
        if self.is_array:
            type_name = type_name + "[]"
        return type_name

    def get_daikon_variable(self):
        lines = [self.var_name(), self.var_kind(), self.dec_type(), self.rep_type(), self.comparability()]
        return u"\n".join([line for line in lines if line != None])

    def __str__(self):
        lines = [self.var_name(), self.var_kind(), self.dec_type(), self.rep_type()]
        return u"\n".join([line for line in lines if line != None])

    def __repr__(self):
        return self.__str__()

class Type:
    Bool                = "bool"
    Int                 = "Int32"
    Float               = "float"
    String              = "string"
    CString             = "cstring"
    PascalString8       = "pascalString8"
    PascalString16      = "pascalString16"
    PascalString32      = "pascalString32"
    Ignore              = "ignore"
    UInt8               = "UInt8"
    UInt16              = "UInt16"
    UInt32              = "UInt32"
    Int16               = "Int16"

    # Type to name map
    name_map = {
        Bool     : "bool",
        Int      : "int",
        Float    : "float",
        String   : "string",
    }

class HachoirToInternalMap:
    # Had to reimport to avoid issues with loading pickle
    import hachoir_core.field
    type_map = {
          hachoir_core.field.bit_field.Bit                  :       Type.Bool,
          hachoir_core.field.bit_field.Bits                 :       Type.Int,
          hachoir_core.field.padding.NullBits               :       Type.Ignore,
          hachoir_core.field.padding.NullBytes              :       Type.Ignore,
          hachoir_core.field.padding.PaddingBytes           :       Type.Ignore,
          hachoir_core.field.integer.UInt8                  :       Type.UInt8,
          hachoir_core.field.integer.UInt16                 :       Type.UInt16,
          hachoir_core.field.integer.UInt32                 :       Type.UInt32,
          hachoir_core.field.integer.Int16                  :       Type.Int16,
          hachoir_core.field.integer.Int32                  :       Type.Int,
          hachoir_core.field.string_field.CString           :       Type.CString,
          hachoir_core.field.string_field.String            :       Type.String,
          hachoir_core.field.string_field.PascalString8     :       Type.PascalString8,
          hachoir_core.field.string_field.PascalString16    :       Type.PascalString16,
          hachoir_core.field.string_field.PascalString32    :       Type.PascalString32,
          hachoir_core.field.byte_field.RawBytes            :       Type.Ignore,
          hachoir_core.field.byte_field.Bytes               :       Type.Ignore,
          hachoir_core.field.bit_field.RawBits              :       Type.Ignore,
          hachoir_core.field.sub_file.SubFile               :       Type.Ignore,
          hachoir_core.field.float.FloatExponent            :       Type.Ignore,
          hachoir_core.field.float.FloatMantissa            :       Type.Ignore,
          hachoir_core.field.link.Link                      :       Type.Ignore,
          }

class Record:
    """
    Record type and value

    Convert field type into an internal representations that can be converted
    into a daikon representation
    """
    # Convert internal types into something we can process
    conv_map = {
        Type.Bool               : lambda x : 1 if x else 0,
        Type.Int                : lambda x : x,
        Type.UInt8              : lambda x : x,
        Type.UInt16             : lambda x : x,
        Type.UInt32             : lambda x : x,
        Type.String             : lambda x : u"\"%s\"" % (repr(x).strip("u").strip("'"), ),
        Type.CString            : lambda x : u"\"%s\"" % (repr(x).strip("u").strip("'"), ),
        Type.PascalString8      : lambda x : u"\"%s\"" % (repr(x).strip("u").strip("'"), ),
        Type.PascalString16     : lambda x : u"\"%s\"" % (repr(x).strip("u").strip("'"), ),
        Type.PascalString32     : lambda x : u"\"%s\"" % (repr(x).strip("u").strip("'"), ),


          Type.Ignore     : lambda x : "nonsensical",
          }

    # Convert internal map into daikon variables. Since Daikon is Java based there's only one
    # integer implementation (i.e., int). Similarly for strings.
    daikon_typename_map = {
            Type.Bool               : "boolean",
            Type.Int                : "int",
            Type.Int16              : "int",
            Type.UInt8              : "int",
            Type.UInt16             : "int",
            Type.UInt32             : "long",
            Type.Float              : "double",
            Type.String             : "string",
            Type.CString            : "string",
            Type.PascalString8      : "string",
            Type.PascalString16     : "string",
            Type.PascalString32     : "string",

            Type.Ignore             : "int",
            }

    def __init__(self, field):
        original_type  = type(field)
        self.original_type  = str(type(field))
        try:
            new_type    = HachoirToInternalMap.type_map[original_type]
        except KeyError:
            assert False, "Cannot map hachoir type %s to internal type" %self.original_type
        self.type          = new_type

        try:
            converter   = self.conv_map[self.type]
            self.value  = converter(field.value)
        except KeyError:
            assert False, "Can't convert python type %s %s" %(self.original_type, self.type)

        self.path             = field.path
        self.name             = field.name
        self.address          = field.address
        self.absolute_address = field.absolute_address
        self.endian           = field.parent.endian
        self.size             = field.size
        self.description      = field.description

        # Record charset for string based classes
        if self.type in [Type.String, Type.CString]:
            self.charset = field.charset
        else:
            self.charset = None

    def get_daikon_type_name(self):
        try:
            return self.daikon_typename_map[self.type]
        except KeyError:
            assert False, "Can't convert %s to daikon" %self.type

    def __str__(self):
        return "%s" %self.value


class StringConstraint:
    ##TODO: Add CString and other types of Strings
    string_types = [Type.String, Type.CString]

    def __init__(self, variable_type, values):
        self.values     =   values
        self.len_max    =   len(max(values, key=len))
        self.len_min    =   len(min(values, key=len))
        self.one_of     =   set(values)
        self.has_uri    =   StringConstraint.contains_uri(values)
        self.encoding   =   StringConstraint.is_str_ascii(values)
        ##TODO: Add character encoding scheme: https://github.com/erikrose/chardet to detect what type of encoding

    def rectify(self, value):
        ## Check for max length seen. Truncate length depending on type (cString with /0)

        ## Map to existing one of?
        if len(value) > self.len_max:
            print "===============> Long string"
        elif value not in self.one_of:
            print "===============> Never before seen value"
        return False, None

    def shiv(self, value, skip_web_shivs=False):
        l = []

        skip_strategies = []
        if skip_web_shivs:
            skip_strategies = ['xss_url_js', 'xss_url_js_null', 
                               'xss_script', 'xss_script_null',
                               'xss_noscript', 'xss_noscript_null',
                               'html_url_jpg', 'html_url_jpg_null',
                               'evil_url', 'evil_url_nonull', 'evil_url_null']
        str_avoider = Avoider(avoider_use_once_deterministic_only, skip_strategies, [])
        str_selector = Chooser(selector_random_avoid_range_breaker)

        fuzzer = StringFuzzer(str(value), {}, [], str_avoider, str_selector)
        result = fuzzer.fuzz()
        while result[1]:
            fuzz_val = result[0]
            if fuzz_val:
                fuzz_val = fuzz_val
            l.append( (fuzz_val, 50) ) 
            result = fuzzer.fuzz()
        return l

    def random_fuzz(self, value, num_fuzz):
        l = []

        str_avoider = Avoider(avoider_nondeterministic_only, [], [])
        str_selector = Chooser(selector_random_avoid_range_breaker)

        fuzzer = StringFuzzer(str(value), {}, [], str_avoider, str_selector)
        result = fuzzer.fuzz()
        count = 0
        while result[1] and count < num_fuzz:
            fuzz_val = result[0]
            if fuzz_val:
                fuzz_val = fuzz_val
            l.append( (fuzz_val, 50) ) 
            result = fuzzer.fuzz()
            count += 1
        return l

    @staticmethod
    def is_str_ascii(l):
        for s in l:
            try:
                s.decode("ascii")
            except:
                return False
        return True

    @staticmethod
    def contains_uri(l):
        uris = [
                'file', 'ftp', 'gopher', "hdl",
                "http", "https", "imap", "mailto", "mms",
                "news", "nntp", "prospero", "rsync", "rtsp", "rtspu", "sftp",
                "shttp", "sip", "sips", "snews", "svn",
                ]
        for uri in uris:
            if any(n.lower().find(uri + ":") > -1 for n in l):
                return True
        return False

    def __str__(self):
        return "%-10d %-10d has_uri:%-12s encoding:%-10s" \
                %(
                    self.len_min,
                    self.len_max,
                    self.has_uri,
                    self.encoding,
                )

class IntegerConstraint:
    type_map = {
        Type.Bool    :   (0,1),
        Type.UInt8   :   (0, 255),
        Type.UInt16  :   (0,65535),
        Type.UInt32  :   (0,4294967295),
        Type.Int     :   (-2147483648,2147483647),
    }
    """
    Rank multipliers
    """
    BOOL_SCORE  = 1
    INT_NEG     = 100
    INT_POS     = 50
    UINT        = 25

    def __init__(self, variable_type, values):
        self.variable_type  = variable_type
        self.values         = values
        # min, max, coverage, negative_value, stats?
        self.vmin           = min(values)
        self.vmax           = max(values)
        self.negative       = any(n<0 for n in values)
        self.coverage       = IntegerConstraint.coverage(values, variable_type)

    @staticmethod
    def coverage(l,t):
        if t not in IntegerConstraint.type_map:
            assert("Could not find type %s in integer type map" %t)
        int_type    =   IntegerConstraint.type_map[t]
        value_range =   float(np.ptp(l))
        type_range  =   float(np.ptp([int_type[0], int_type[1]]))
        #print value_range, type_range, type_range-value_range
        if type_range - value_range == 0.0:
            return 100.0
        return (value_range / type_range) * 100.0

    def shiv(self, value):
        int_type = IntegerConstraint.type_map[self.variable_type]
        l = [(int_type[0], 100),
             (int_type[1], 100)]
        if value - 1 > int_type[0]:
            l.append((value - 1, 75))
        if value + 1 < int_type[1]:
            l.append((value + 1, 75))
        if int_type[0] < 0:
            l += [(0, 75), 
                  (-1, 75)]
        return l

    def random_fuzz(self, value):
        '''
        Generate random value that lies outside the comfortzone.
        '''
        ## If we have full coverage, skip

        int_type = IntegerConstraint.type_map[self.variable_type]

        ##TODO: Adjust rank score by coverage of learning
        if self.variable_type == Type.Bool:
            return value, IntegerConstraint.BOOL_SCORE

        ## If we have a signed number, flip a coin to decide whether to try negative or positive value
        if self.variable_type == Type.Int:
            if random.randint(0,1) == 0:
                # pick from the smallest number we've seen to the smallest possible
                fuzz_val    = random.randrange(self.vmin, int_type[0])
                rank        = IntegerConstraint.INT_NEG
            else:
                fuzz_val    = random.randrange(self.vmax, int_type[1])
                rank        = IntegerConstraint.INT_POS
        else:
            fuzz_val        = random.randrange(self.vmax, int_type[1])
            rank            = IntegerConstraint.UINT
        
        rank = rank * (1.0 - (self.coverage/100.0))

        return fuzz_val, rank

    def rectify(self, value):
        ## Negative value
        if self.negative and value < 0:
            print "Rectifying negative value:%d to 0" %value
            return True, 0

        ## Truncate
        if value > self.vmax:
            print "Rectifying value:%d to maximum:%d" %(value, self.vmax)
            return True, self.vmax

        ##TODO: SHould I rectify vmin??
        return False, None

    def __str__(self):
        # type, vmin, vmax, coverage%, negative
        return "%-10d %-10d %-20f %-10s"\
                %(
                    #self.variable_type,
                    self.vmin,
                    self.vmax,
                    self.coverage,
                    self.negative
                )
